#ifndef _BSTRING_CPP
#define _BSTRING_CPP

#include <string>
#include <new>
#include <algorithm>

#include "BString.h"

using namespace std;

struct BString::BStringImpl
{
   enum { bsIncr = 16 };
   char*  bsData; 
   size_t bsLen; 
   size_t bsReserved;

   BStringImpl( const char* rhs="" );
   BStringImpl( const BStringImpl& rhs );
   ~BStringImpl();
   BStringImpl& operator=( const BStringImpl& rhs );
   void swap( BStringImpl& rhs );
   
   static size_t cStrLen( const char* cs );
   static int stringCmpr( const char* lhs, const char* rhs, int n );
}; // BString::BStringImpl

//================================================================

BString::BStringImpl::BStringImpl( const char* rhs )
{
   bsLen = cStrLen( rhs );
   bsReserved = ((bsLen+1)/bsIncr+1)*bsIncr;

   bsData = new char[bsReserved];
   for ( int k(0); k < bsLen; ++k )  bsData[k] = rhs[k]; 
} // BString::BStringImpl::BStringImpl

BString::BStringImpl::BStringImpl( const BString::BStringImpl& rhs )
{
   bsData = new char[rhs.bsReserved];
   bsReserved = rhs.bsReserved;
   for ( int k(0); k < rhs.bsLen; ++k )  bsData[k] = rhs.bsData[k]; 
   bsLen = rhs.bsLen;
} // BString::BStringImpl::BStringImpl

BString::BStringImpl::~BStringImpl() {  delete bsData;  }

BString::BStringImpl& BString::BStringImpl::operator=( const BString::BStringImpl& rhs )
{
   BString::BStringImpl tmp( rhs );
   if ( &rhs != this )
      swap( tmp );
   return *this;
} // BString::BStringImpl::BStringImpl& operator=

void BString::BStringImpl::swap( BString::BStringImpl& rhs )
{
   std::swap( bsLen, rhs.bsLen );
   std::swap( bsReserved, rhs.bsReserved );
   std::swap( bsData, rhs.bsData );
} // BString::BStringImpl::swap

//=====================   ======================

size_t BString::BStringImpl::cStrLen( const char cs[] )
{
   const char* p( cs );
   while ( *p != '\0' )  ++p;
   return p-cs;
} // BString::BStringImpl::cStrLen

int BString::BStringImpl::stringCmpr( const char* lhs, const char* rhs, int n )
{
   int k( 0 );
   while ( lhs[k] == rhs[k] && k < n-1 ) ++k;
   return ( static_cast<int>(lhs[k]) - static_cast<int>(rhs[k]) );
} // BString::BStringImpl::stringCmpr

//================================================================

BString::BString( const char* rhs )
{
   try {  bsHandle = new BStringImpl( rhs );  }
   catch( bad_alloc& )
   {  throw BStringXptn( "BString", BStringXptn::allocFailed );  }
} // BString::BString

BString::BString( const BString& rhs )
{
   try {  bsHandle = new BStringImpl( *(rhs.bsHandle) );  }
   catch( bad_alloc& )
   {  throw BStringXptn( "BString", BStringXptn::allocFailed );  }
} // BString::BString

BString::~BString() {  delete bsHandle;  }
/*
BString::BString( const char* rhs )
{
   bsHandle->bsLen = BStringImpl::cStrLen( rhs );
   bsHandle->bsReserved = ((bsHandle->bsLen+1)/BStringImpl::bsIncr+1)*BStringImpl::bsIncr;

   try {  bsHandle->bsData = new char [bsHandle->bsReserved];  }
   catch( bad_alloc )
   {  throw BStringXptn( "BString", BStringXptn::allocFailed );  }
   for ( int k(0); k < bsHandle->bsLen; ++k )  bsHandle->bsData[k] = rhs[k]; 
} // BString::BString

BString::BString( const BString& rhs )
{
   bsHandle->bsReserved = (rhs.bsHandle)->bsReserved;

   try {  bsHandle->bsData = new char [bsHandle->bsReserved];  }
   catch( bad_alloc )
   {  throw BStringXptn( "BString", BStringXptn::allocFailed );  }
   for ( int k(0); k < (rhs.bsHandle)->bsLen; ++k )  bsHandle->bsData[k] = (rhs.bsHandle)->bsData[k];
   bsHandle->bsLen = (rhs.bsHandle)->bsLen;
} // BString::BString
*/
const char* BString::c_str() const
{  bsHandle->bsData[bsHandle->bsLen] = '\0';  return bsHandle->bsData;  }

size_t BString::length() const {  return bsHandle->bsLen;  }

bool BString::empty() const {  return ( bsHandle->bsLen == 0 );  }

char& BString::operator[]( int pos ) const {  return bsHandle->bsData[pos];  }

char& BString::at( int k ) const
{
   if ( k < 0 || bsHandle->bsLen <= k )
      throw BStringXptn( "at", BStringXptn::outOfRange, k );
   return bsHandle->bsData[k];
} // BString::at

BString& BString::operator+=( const BString& rhs )
{
   if ( bsHandle->bsLen + (rhs.bsHandle)->bsLen + 1 > bsHandle->bsReserved )
   {
      char* tmp;
      size_t tmpRes( ((bsHandle->bsLen+(rhs.bsHandle)->bsLen+1)/BStringImpl::bsIncr+1)*BStringImpl::bsIncr );
      try {  tmp = new char[tmpRes];  }
      catch( bad_alloc )
      {  throw BStringXptn( "operator+=",
                            BStringXptn::allocFailed );  }
      for ( int k(0); k < bsHandle->bsLen; ++k )  tmp[k] = bsHandle->bsData[k];
      delete[] bsHandle->bsData;
      bsHandle->bsData = tmp;
   }
   for ( int j(0), k(bsHandle->bsLen); j < (rhs.bsHandle)->bsLen; ++j, ++k )
      bsHandle->bsData[k] = (rhs.bsHandle)->bsData[j];
   bsHandle->bsLen += (rhs.bsHandle)->bsLen;
   return *this;
} // BString::operator+=

BString& BString::append( const BString& rhs ) { return (*this += rhs); }
/*
BString& BString::append( const BString& rhs )
{
   if ( bsHandle->bsLen + (rhs.bsHandle)->bsLen + 1 > bsHandle->bsReserved )
   {
      char* tmp;
      size_t tmpRes( ((bsHandle->bsLen+(rhs.bsHandle)->bsLen+1)/bsHandle->bsIncr+1)*bsHandle->bsIncr );
      try {  tmp = new char[tmpRes];  }
      catch( bad_alloc )
      {  throw BStringXptn( "append",
                            BStringXptn::allocFailed );  }
      for ( int k(0); k < bsHandle->bsLen; ++k )  tmp[k] = bsHandle->bsData[k];
      delete[] bsHandle->bsData;
      bsHandle->bsData = tmp;
   }
   for ( int j(0), k(bsHandle->bsLen); j < (rhs.bsHandle)->bsLen; ++j, ++k )
      bsHandle->bsData[k] = (rhs.bsHandle)->bsData[j];
   bsHandle->bsLen += (rhs.bsHandle)->bsLen;
   return *this;
} // BString::append
/*
BString& BString::operator=( const BString& rhs )
{
   if ( &rhs != this )  // if ( ! )
   {
   //     
      char* tmp;
      try {  tmp = new char[(rhs.bsHandle)->bsReserved];  }
      catch( bad_alloc& )
      { throw BStringXptn( "operator=", BStringXptn::allocFailed ); }

      delete[] bsHandle->bsData;
   //     rhs
      bsHandle->bsData = tmp;
      bsHandle->bsReserved = (rhs.bsHandle)->bsReserved; 
      for ( int k(0); k < (rhs.bsHandle)->bsLen; ++k )
         bsHandle->bsData[k] = (rhs.bsHandle)->bsData[k];
      bsHandle->bsLen = (rhs.bsHandle)->bsLen;
   }
   return *this;
} // BString::operator=
*/
BString& BString::operator=( const BString& rhs )
{
   if ( &rhs != this )
   {
      try {  BString tmp( rhs );
             swap( tmp );  }
      catch( BStringXptn& x )
      {  strcpy( x.funcName, "operator=" );
         throw;  }
   }
   return *this; 
} // BString::operator=

BString& BString::assign( const BString& rhs )
{ return (*this = rhs); }

void BString::swap( BString& rhs )
{
   std::swap( bsHandle, rhs.bsHandle );
/*   BStringImpl* sv( rhs.bsHandle );
   rhs.bsHandle = bsHandle;  bsHandle = sv; */
} // BString::swap

int BString::compare( const BString& rhs ) const
{
   int fv( BStringImpl::stringCmpr(bsHandle->bsData, (rhs.bsHandle)->bsData,
                      min(bsHandle->bsLen, (rhs.bsHandle)->bsLen)) );
   if ( fv == 0 )  fv = bsHandle->bsLen - (rhs.bsHandle)->bsLen;
   return fv;
} // BString::compare
/*
BString& BString::assign( const BString& rhs )
{
   if ( &rhs != this )  // if ( ! )
   {
   //     
      char* tmp;
      try {  tmp = new char[(rhs.bsHandle)->bsReserved];  }
      catch( bad_alloc& )
      {  throw BStringXptn( "assign", BStringXptn::allocFailed );  }

      delete[] bsHandle->bsData;
   //     rhs
      bsHandle->bsData = tmp;
      bsHandle->bsReserved = (rhs.bsHandle)->bsReserved; 
      for ( int k(0); k < (rhs.bsHandle)->bsLen; ++k )
         bsHandle->bsData[k] = (rhs.bsHandle)->bsData[k];
      bsHandle->bsLen = (rhs.bsHandle)->bsLen;
   }
   return *this;
} // BString::assign
*/
//=====================   =======================

ostream& operator<<( ostream& tout, const BString& rhs )
{
   for ( int k(0); k < (rhs.bsHandle)->bsLen; ++k )
      tout << (rhs.bsHandle)->bsData[k];
   return tout;
} // operator<<( ofstream, BString

bool operator==( const BString& lhs, const BString& rhs )
{
   int fv( BString::BStringImpl::stringCmpr((lhs.bsHandle)->bsData,
                                            (rhs.bsHandle)->bsData,
             min((lhs.bsHandle)->bsLen, (rhs.bsHandle)->bsLen)) );
   if ( fv == 0 )  fv = (lhs.bsHandle)->bsLen - (rhs.bsHandle)->bsLen;
   return ( fv == 0 );
} // operator==

BString operator+( const BString& lhs, const BString& rhs )
{
   BString fv( lhs );
   fv += rhs;
   return fv;
} // BStribg operator+

#endif // _BSTRING_CPP
